require 'ams_CopyAlongCurve.rb'

module AMS
  class CopyAlongCurve

    CURSORS = {
      :verify       => [1,9],
      :select_curve => [1,6],
      :select_group => [1,9]
    }
    dir = File.dirname(__FILE__)
    CURSORS.each { |name, origin|
      origin[2] = UI.create_cursor(File.join(dir, name.to_s + '.png'), origin[0], origin[1])
    }

    @@spacing = 10.to_l
    @@align = true

    def initialize
      @step = 0
      @hovered_edge = nil
      @hovered_edge_tra = nil
      @group = nil
      @group_definition = nil
      @loc_group_normal_tra = nil
      @group_transformations = {}
      @group_bb = []
      @normals = {}
      @vertices = []
      @tra_scale = 1
      @curve_length = nil
      @starting_distance = 0
      @initial_index = 0
      @final_index = 0
      @cursor_id = CURSORS[:select_group][2]
    end

    def transformation_from_pin_dir(point, vector)
      Geom::Transformation.new(point, vector)
    end

    def rotate_transformation_zaxis_to_dir(matrix, dir, new_point)
      dir1 = matrix.zaxis.normalize
      dir2 = dir.normalize
      cos_theta = dir1.dot(dir2)
      normal = nil
      if cos_theta.abs > 0.9999995
        normal = matrix.xaxis
      else
        normal = dir1.cross(dir2)
      end
      normal.normalize!
      yaxis1 = normal.cross(dir1).normalize
      yaxis2 = normal.cross(dir2).normalize
      t1 = Geom::Transformation.new(dir1, yaxis1, normal, ORIGIN)
      t2 = Geom::Transformation.new(dir2, yaxis2, normal, ORIGIN)
      t3 = t2 * (t1.inverse * matrix)
      Geom::Transformation.new(t3.xaxis, t3.yaxis, t3.zaxis, new_point)
    end

    def scale_vector(v, s)
      Geom::Vector3d.new(v.x * s, v.y * s, v.z * s)
    end

    def draw_matrix(view, tra, scale)
      c = tra.origin
      view.line_stipple = ''
      view.line_width = 2
      view.drawing_color = ::Sketchup::Color.new(255, 10, 10)
      view.draw_line(c, c + scale_vector(tra.xaxis, scale))
      view.drawing_color = ::Sketchup::Color.new(10, 255, 10)
      view.draw_line(c, c + scale_vector(tra.yaxis, scale))
      view.drawing_color = ::Sketchup::Color.new(10, 10, 255)
      view.draw_line(c, c + scale_vector(tra.zaxis, scale))
    end

    def calculate_normals
      edges = {}
      @normals.clear
      # Get all consequent edge normals
      edge = @hovered_edge
      v1 = edge.start
      v2 = edge.end
      p1 = v1.position.transform(@hovered_edge_tra)
      p2 = v2.position.transform(@hovered_edge_tra)
      dp = p2 - p1
      @normals[0] = [transformation_from_pin_dir(p1, dp.normalize), dp.length]
      edges[@hovered_edge] = 0
      count = 1
      while true
        next_edge = v2.edges[0]
        if next_edge == edge
          if v2.edges[1]
            next_edge = v2.edges[1]
          else
            break
          end
        end
        break if edges.has_key?(next_edge)
        v1 = v2
        v2 = next_edge.other_vertex(v1)
        p1 = p2
        p2 = v2.position.transform(@hovered_edge_tra)
        dp = p2 - p1
        @normals[count] = [rotate_transformation_zaxis_to_dir(@normals[count - 1][0], dp.normalize, p1), dp.length]
        edges[next_edge] = count
        edge = next_edge
        count += 1
      end
      # Get all preceding edge normals
      edge = @hovered_edge
      v1 = edge.start
      v2 = nil
      p1 = v1.position.transform(@hovered_edge_tra)
      p2 = nil
      count = -1
      while true
        next_edge = v1.edges[0]
        if next_edge == edge
          if v1.edges[1]
            next_edge = v1.edges[1]
          else
            break
          end
        end
        break if edges.has_key?(next_edge)
        v2 = next_edge.other_vertex(v1)
        p2 = v2.position.transform(@hovered_edge_tra)
        dp = p1 - p2
        @normals[count] = [rotate_transformation_zaxis_to_dir(@normals[count + 1][0], dp.normalize, p2), dp.length]
        edges[next_edge] = count
        edge = next_edge
        v1 = v2
        p1 = p2
        count -= 1
      end
      # Get vertices and calculate length
      @curve_length = 0
      @initial_index = 0
      @final_index = 0
      @vertices = @normals.keys.sort.map { |index|
        @curve_length += @normals[index][1]
        @initial_index = index if index < @initial_index
        @final_index = index if index > @final_index
        @normals[index][0].origin
      }
      last_data = @normals[@final_index]
      @vertices << Geom::Point3d.new(0, 0, last_data[1]).transform(last_data[0])
      @curve_length = @curve_length.to_l
      @loc_group_normal_tra = (@normals[0][0].inverse * @group.transformation).to_a
      # Compute starting distance
      @starting_distance = @loc_group_normal_tra[14]
      for i in @initial_index...0
        @starting_distance += @normals[i][1]
      end
    end

    def calculate_group_transformation_on_curve(desired_distance)
      index = @initial_index
      current_distance = 0
      while @normals.has_key?(index)
        normal_data = @normals[index]
        if current_distance + normal_data[1] > desired_distance
          @loc_group_normal_tra[14] = desired_distance - current_distance
          res = normal_data[0] * Geom::Transformation.new(@loc_group_normal_tra)
          if @@align
            return normal_data[0] * Geom::Transformation.new(@loc_group_normal_tra)
          else
            origin = Geom::Point3d.new(@loc_group_normal_tra[12..14]).transform(normal_data[0])
            gmatrix = @group.transformation.to_a
            gmatrix[12] = origin.x
            gmatrix[13] = origin.y
            gmatrix[14] = origin.z
            return Geom::Transformation.new(gmatrix)
          end
        end
        current_distance += normal_data[1]
        index += 1
      end
      nil
    end

    def calculate_group_transformations
      @group_transformations.clear
      distance = @starting_distance.to_f
      while true
        distance -= @@spacing.to_f
        break if distance < 0
        tra = calculate_group_transformation_on_curve(distance)
        break unless tra
        @group_transformations[tra] = @group_bb.map { |pt| pt.transform(tra) }
      end
      distance = @starting_distance.to_f
      while true
        distance += @@spacing.to_f
        tra = calculate_group_transformation_on_curve(distance)
        break unless tra
        @group_transformations[tra] = @group_bb.map { |pt| pt.transform(tra) }
      end
    end

    def copy_along_curve
      model = ::Sketchup.active_model
      if ::Sketchup.version.to_i > 6
        model.start_operation('Copy Along Curve', true)
      else
        model.start_operation('Copy Along Curve')
      end
      @group_transformations.each { |tra, inst_bb|
        inst = model.active_entities.add_instance(@group_definition, tra)
        inst.material = @group.material
        inst.layer = @group.layer
        inst.name = @group.name
        inst.casts_shadows = @group.casts_shadows?
        inst.receives_shadows = @group.receives_shadows?
        dicts = @group.attribute_dictionaries
        if dicts
          dicts.each { |dict|
            dict.each { |k, v|
              inst.set_attribute(dict.name, k, v)
            }
          }
        end
      }
      model.commit_operation
    rescue Exception => err
      model.abort_operation
      ::UI.messagebox(err)
      puts err
      puts err.backtrace.join("\n")
    end

    def update_status
      if @step == 0
        ::Sketchup.set_status_text("选择要阵列的目标群组/组件。", SB_PROMPT)
      elsif @step == 1
        ::Sketchup.set_status_text("选择最靠近目标群组/组件的那条边线。", SB_PROMPT)
      elsif @step == 2
        ::Sketchup.set_status_text("调整设置或单击鼠标左键完成阵列。    曲线长度: #{@curve_length}    阵列间距: #{@@spacing}    阵列数量: #{@group_transformations.size}个    对齐方式 (← → 方向键): #{@@align ? '是' : '否'}", SB_PROMPT)
      end
      ::Sketchup.set_status_text("请输入阵列间距", SB_VCB_LABEL)
      ::Sketchup.set_status_text(@@spacing.to_s, SB_VCB_VALUE)
      onSetCursor
    end

    def activate
      model = ::Sketchup.active_model
      opts = model.rendering_options
      model.selection.clear
      @show_edges = opts['EdgeDisplayMode']
      @show_profiles = opts['DrawSilhouettes']
      update_status
    end

    def deactivate(view)
      model = ::Sketchup.active_model
      model.selection.clear
      opts = model.rendering_options
      opts['EdgeDisplayMode'] = @show_edges
      opts['DrawSilhouettes'] = @show_profiles
      view.invalidate
    end

    def resume(view)
      update_status
      view.invalidate
    end

    def onMouseEnter(view)
      update_status
      view.invalidate
    end

    def onUserText(text, view)
      begin
        if text[-1] == 'c'
          num_copies = text.to_i.abs
          @@spacing = (@curve_length.to_f / (num_copies + 1)).to_l
        else
          @@spacing = text.to_l
          @@spacing = 0.1.mm if @@spacing < 0.1.mm
        end
        calculate_group_transformations if @step == 2
        update_status
        view.invalidate
      rescue Exception => e
        update_status
      end
    end

    def onKeyDown(key, repeat, flags, view)
      return if repeat != 1
      prev_align = @@align
      if key == VK_LEFT
        @@align = false
      elsif key == VK_RIGHT
        @@align = true
      end
      if @@align != prev_align && @step == 2
        calculate_group_transformations
        update_status
        view.invalidate
      end
    end

    def onMouseMove(flags, x, y, view)
      if @step == 1
        ph = view.pick_helper(x, y)
        entity = ph.leaf_at(0)
        edge = nil
        if entity.is_a?(::Sketchup::Edge)
          edge = entity
        end
        if edge != @hovered_edge
          @hovered_edge = edge
          @hovered_edge_tra = ph.transformation_at(0)
          view.invalidate
        end
      end
    end

    def onLButtonUp(flags, x, y, view)
      model = ::Sketchup.active_model
      if @step == 0
        ph = view.pick_helper(x, y)
        entity = ph.best_picked
        if entity.is_a?(::Sketchup::Group) || entity.is_a?(::Sketchup::ComponentInstance)
          @group = entity
          @tra_scale = @group.bounds.diagonal * 0.1
          if @group.is_a?(::Sketchup::ComponentInstance)
            @group_definition = @group.definition
          elsif ::Sketchup.version.to_i >= 14
            @group_definition = @group.entities.parent
          else
            @group_definition = @group.model.definitions.find { |d|
              d.group? && d.instances.include?(@group)
            }
          end
          a = @group_definition.bounds.min
          b = @group_definition.bounds.max
          @group_bb = [
            Geom::Point3d.new(a.x, a.y, a.z), Geom::Point3d.new(b.x, a.y, a.z),
            Geom::Point3d.new(b.x, a.y, a.z), Geom::Point3d.new(b.x, b.y, a.z),
            Geom::Point3d.new(b.x, b.y, a.z), Geom::Point3d.new(a.x, b.y, a.z),
            Geom::Point3d.new(a.x, b.y, a.z), Geom::Point3d.new(a.x, a.y, a.z),
            Geom::Point3d.new(a.x, a.y, b.z), Geom::Point3d.new(b.x, a.y, b.z),
            Geom::Point3d.new(b.x, a.y, b.z), Geom::Point3d.new(b.x, b.y, b.z),
            Geom::Point3d.new(b.x, b.y, b.z), Geom::Point3d.new(a.x, b.y, b.z),
            Geom::Point3d.new(a.x, b.y, b.z), Geom::Point3d.new(a.x, a.y, b.z),
            Geom::Point3d.new(a.x, a.y, a.z), Geom::Point3d.new(a.x, a.y, b.z),
            Geom::Point3d.new(b.x, a.y, a.z), Geom::Point3d.new(b.x, a.y, b.z),
            Geom::Point3d.new(b.x, b.y, a.z), Geom::Point3d.new(b.x, b.y, b.z),
            Geom::Point3d.new(a.x, b.y, a.z), Geom::Point3d.new(a.x, b.y, b.z)
          ]
          @step = 1
          model.selection.clear
          model.selection.add(entity)
          opts = model.rendering_options
          opts['EdgeDisplayMode'] = @show_edges
          opts['DrawSilhouettes'] = @show_profiles
          @cursor_id = CURSORS[:select_curve][2]
          update_status
        end
      elsif @step == 1
        if @hovered_edge
          @step = 2
          calculate_normals
          calculate_group_transformations
          opts = model.rendering_options
          opts['EdgeDisplayMode'] = false
          opts['DrawSilhouettes'] = false
          @cursor_id = CURSORS[:verify][2]
          update_status
        end
      elsif @step == 2
        copy_along_curve
        @step = 0
        @hovered_edge = nil
        @hovered_edge_tra = nil
        @group = nil
        @group_definition = nil
        @loc_group_normal_tra = nil
        @group_transformations.clear
        @group_bb.clear
        @normals.clear
        @vertices.clear
        @curve_length = nil
        @initial_index = 0
        @final_index = 0
        model.selection.clear
        opts = model.rendering_options
        opts['EdgeDisplayMode'] = @show_edges
        opts['DrawSilhouettes'] = @show_profiles
        @cursor_id = CURSORS[:select_group][2]
        update_status
      end
    end

    def draw(view)
      if @step == 1
        if @hovered_edge
          view.drawing_color = ::Sketchup::Color.new(0, 230, 0)
          view.line_width = 4
          view.line_stipple = ''
          pt1 = @hovered_edge.start.position.transform(@hovered_edge_tra)
          pt2 = @hovered_edge.end.position.transform(@hovered_edge_tra)
          view.draw(GL_LINES, [pt1, pt2])
        end
      elsif @step == 2
        @group_transformations.each { |tra, inst_bb|
          #draw_matrix(view, tra, @tra_scale)
          view.drawing_color = ::Sketchup::Color.new(159, 10, 209)
          view.line_width = 1
          view.line_stipple = ''
          view.draw(GL_LINES, inst_bb)
        }
        view.drawing_color = ::Sketchup::Color.new(0, 255, 29)
        view.line_width = 2
        view.line_stipple = ''
        view.draw(GL_LINE_STRIP, @vertices)
      end
    end

    def getExtents
      model = ::Sketchup.active_model
      bb = Geom::BoundingBox.new
      if ::Sketchup.version.to_i > 6
        model.entities.each { |e|
          next if e.is_a?(::Sketchup::Text)
          if e.visible? && e.layer.visible?
            bb.add(e.bounds)
          end
        }
      end
      bb.add(@vertices) if !@vertices.empty?
      bb.empty? ? model.bounds : bb
    end

    def onSetCursor
      ::UI.set_cursor(@cursor_id)
    end

  end # class CopyAlongCurve
end # module AMS

unless file_loaded?(__FILE__)
  UI.menu("Plugins").add_item("曲线阵列") {
    ::Sketchup.active_model.select_tool(AMS::CopyAlongCurve.new)
  }
  file_loaded(__FILE__)
end

    # 以下部分是添加图标的代码
    toolbar = UI::Toolbar.new "曲线阵列"
    cmd = UI::Command.new("曲线阵列") {
    ::Sketchup.active_model.select_tool(AMS::CopyAlongCurve.new)
  }
    cmd.tooltip = "曲线阵列"
    cmd.status_bar_text = '沿着曲线/边缘路径阵列复制组和组件。'
    cmd.small_icon = "icon.png"
    cmd.large_icon = "icon.png"
    toolbar = toolbar.add_item cmd
    toolbar.show
    puts cmd.large_icon
    # 以上部分是添加图标的代码
